//! Validation for various kinds of git related items.
#![deny(missing_docs, rust_2018_idioms)]
#![forbid(unsafe_code)]

///
pub mod reference;

///
pub mod tag;

///
pub mod submodule;

///
pub mod path;
